# ZpCron - Cron-like scheduler for Zope
# Copyright (C) 2008  Jan Janech <janech@kios.sk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import crontab
import timer

import Globals
import Acquisition
import AccessControl
import OFS.Traversable
import App.Management

from AccessControl.SecurityManagement import newSecurityManager, getSecurityManager
from ZPublisher.BaseRequest import RequestContainer, BaseRequest
from Zope2.App.startup import zpublisher_exception_hook #!!!
# import Zope2

import cStringIO
import urllib
import sys
import os

lines = None
rebootlines = None
restart = True

timerObject = None

application = None

class ZpCron(Globals.Persistent, Acquisition.Implicit, OFS.Traversable.Traversable, App.Management.Tabs, crontab.CronTab):
    meta_type = "ZpCron"
    
    manage_options = (
        {'label': 'CronTab', 'action': 'manage_workspace'},
        {'label': 'Timer', 'action': 'manage_timer'},
    )
        
    __ac_permissions__ = (
        ('View management screens', ('manage_workspace', 'manage_edit', 'manage_restart', 'manage_timer'), ('Manager',)),
    )
    
    id = 'ZpCron'
    title = 'CronTab'
    
    icon='misc_/ZpCron/clock'
    
    # management
    manage_workspace = Globals.DTMLFile('www/edit', globals())
    manage_timer = Globals.DTMLFile('www/timer', globals())
    
    def __init__(self):
        self.crontab = ''
        self.update()
    
    def title_or_id(self):
        return self.title
    
    def manage_restart(self, REQUEST = None):
        """ """
        self.restart()
        return self.manage_workspace(REQUEST = REQUEST)
    
    def manage_edit(self, crontab, REQUEST = None):
        """ """
        self.crontab = crontab
        self.update()
        return self.manage_workspace(REQUEST = REQUEST)
    
    def update(self):
        global timerObject
        
        self.reset()
        self.parseFile(cStringIO.StringIO(self.crontab))
        if timerObject is not None:
            timerObject.change()
    
    # execution
    @staticmethod
    def tick():
        # global application
        
        # if application is None:
        #     application = zapp = Zope2.app()
        # else:
        #     zapp = application
        from Zope2 import DB
        conn = DB.open()
        
        app = conn.root()['Application'].__of__(RequestContainer(REQUEST = BaseRequest(os.environ.copy())))
        
        self = app.Control_Panel.ZpCron
        
        cmds = []
        for waittime, cmd in self.getFirstEvents():
            if waittime <= 0:
                cmds.append(cmd)
        if cmds:
            self.refresh()
        ret = []
        oldUser = None
        for user, url, params in cmds:
            if user != oldUser:
                userObj = app.acl_users.getUser(user).__of__(app.acl_users)
                newSecurityManager(None, userObj)
                app.REQUEST['AUTHENTICATED_USER'] = userObj
                oldUser = user
            app.REQUEST['URL'] = url
            obj = app.restrictedTraverse(url)
            output = None
            try:
                output = obj(**dict(params))
            except:
                exc, val, trac = sys.exc_info()
                try:
                    output = zpublisher_exception_hook(obj, app.REQUEST, exc, val, trac)
                except:
                    pass
            if output:
                ret.append(output)
        
        conn.close()
        
        return self.getFirstEvents()[0][0]
    
    # timer thread
    def start(self):
        global timerObject
        
        if timerObject is None:
            timerObject = timer.Timer(self.tick)
            timerObject.start()
    
    def stop(self):
        global timerObject
        
        if timerObject is not None:
            timerObject.stop()
            timerObject = None
    
    def restart(self):
        global timerObject
        
        if timerObject is None:
            self.start()
        else:
            timerObject.restart()
    
    def timerRunning(self):
        global timerObject
        
        return timerObject is not None and timerObject.running()
    
    def timerLastEvent(self):
        global timerObject
        
        if timerObject is None:
            return None
        else:
            return timerObject.lastTime()
    
    # crontab overrides
    def reset(self):
        global lines, rebootlines
        
        lines = []
        rebootlines = []
    
    def getRestart(self):
        global restart
        
        r = restart
        restart = False
        return r
    
    def getLines(self):
        return lines
    
    def getRebootLines(self):
        return rebootlines
    
    def postparseCommand(self, command, USER):
        assert command[0] == '/'
        
        x = command.split('?', 1)
        if len(x) > 1:
            obj, params = x
        else:
            obj = x[0]
            params = ""
        if params:
            params = [(name, urllib.unquote(value)) for name, value in (i.split('=', 1) for i in params.split('&'))]
        else:
            params = []
        return USER, obj, params

Globals.InitializeClass(ZpCron)
